@@ -9,16 +9,15 @@
 #    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 #    License for the specific language governing permissions and limitations
 #    under the License.
-from __future__ import absolute_import
 
 import os
 
 from gabbi import fixture
 import os_resource_classes as orc
+import os_traits as ot
 from oslo_config import cfg
 from oslo_config import fixture as config_fixture
 from oslo_log.fixture import logging_error
-from oslo_middleware import cors
 from oslo_policy import opts as policy_opts
 from oslo_utils.fixture import uuidsentinel as uuids
 from oslo_utils import uuidutils
@@ -37,7 +36,7 @@ from placement.tests.functional.fixtures import capture
 from placement.tests.unit import policy_fixture
 
 
-# This global conf is not a global olso_config.cfg.CONF. It's a global
+# This global conf is not a global oslo_config.cfg.CONF. It's a global
 # used locally to work around a limitation in the way that gabbi instantiates
 # the WSGI application being tested.
 CONF = None
@@ -51,6 +50,9 @@ def setup_app():
 class APIFixture(fixture.GabbiFixture):
     """Setup the required backend fixtures for a basic placement service."""
 
+    # TODO(stephenfin): Remove this once we drop the deprecated policy rules
+    _secure_rbac = False
+
     def start_fixture(self):
         global CONF
         # Set up stderr and stdout captures by directly driving the
@@ -73,18 +75,23 @@ class APIFixture(fixture.GabbiFixture):
         self.conf_fixture.setUp()
         conf.register_opts(self.conf_fixture.conf)
         self.conf_fixture.config(group='api', auth_strategy='noauth2')
+        self.conf_fixture.config(
+            group='oslo_policy',
+            enforce_scope=self._secure_rbac,
+            enforce_new_defaults=self._secure_rbac,
+        )
 
         self.placement_db_fixture = fixtures.Database(
             self.conf_fixture, set_config=True)
         self.placement_db_fixture.setUp()
 
         self.context = context.RequestContext()
+        # Some database interaction methods require access to the oslo config
+        # via the context. Within the WSGI application this is taken care of
+        # but here in the fixtures we use some of those methods to create
+        # entities.
+        self.context.config = self.conf_fixture.conf
 
-        # Register CORS opts, but do not set config. This has the
-        # effect of exercising the "don't use cors" path in
-        # deploy.py. Without setting some config the group will not
-        # be present.
-        self.conf_fixture.register_opts(cors.CORS_OPTS, 'cors')
         # Set default policy opts, otherwise the deploy module can
         # NoSuchOptError.
         policy_opts.set_defaults(self.conf_fixture.conf)
@@ -360,9 +367,244 @@ class NUMAAggregateFixture(APIFixture):
             tb.set_traits(ss, 'MISC_SHARES_VIA_AGGREGATE')
 
 
+class NUMANetworkFixture(APIFixture):
+    """An APIFixture representing compute hosts with characteristics such as:
+
+    * A root compute node provider with no resources (VCPU/MEMORY_MB in NUMA
+      providers, DISK_GB provided by a sharing provider).
+    * NUMA nodes, providing VCPU and MEMORY_MB resources (with interesting
+      min_unit and step_size values), decorated with the HW_NUMA_ROOT trait.
+    * Each NUMA node is associated with some devices.
+    * Two network agents, themselves devoid of resources, parenting different
+      kinds of network devices.
+
+    * A more "normal" compute node provider with VCPU/MEMORY_MB/DISK_GB
+      resources.
+    * Some NIC subtree roots, themselves devoid of resources, decorated with
+      the HW_NIC_ROOT trait, parenting PF providers, on different physical
+      networks, with VF resources.
+
+                          +-----------------------+
+                          | sharing storage (ss1) |
+                          | MISC_SHARES_VIA_AGG.. |
+                          |   DISK_GB:2000        |............(to cn2)......>>
+                          |   agg: [aggA]         |
+                          +-----------+-----------+
+                                      :
+                          +-----------------------------+
+                          |     compute node (cn1)      |
+                          | COMPUTE_VOLUME_MULTI_ATTACH |
+                          |      (no inventory)         |
+                          |        agg: [aggA]          |
+                          +---------------+-------------+
+                                          |
+              +--------------------+------+--------+-------------------+
+              |                    |               |                   |
+    +---------+--------+ +---------+--------+      |                   |
+    | numa0            | | numa1            |      |       +-----------+------+
+    | HW_NUMA_ROOT     | | HW_NUMA_ROOT, FOO|      |       | ovs_agent        |
+    | VCPU: 4 (2 used) | | VCPU: 4          |      |       | VNIC_TYPE_NORMAL |
+    | MEMORY_MB: 2048  | | MEMORY_MB: 2048  |      |       +-----------+------+
+    |   min_unit: 512  | |   min_unit: 256  |      |                   |
+    |   step_size: 256 | |   max_unit: 1024 |      |                   |
+    +---+----------+---+ +---+----------+---+  +---+--------------+    |
+        |          |         |          |      | sriov_agent      |    |
+    +---+---+  +---+---+ +---+---+  +---+---+  | VNIC_TYPE_DIRECT |    |
+    |fpga0  |  |pgpu0  | |fpga1_0|  |fpga1_1|  +---+--------------+    |
+    |FPGA:1 |  |VGPU:8 | |FPGA:1 |  |FPGA:1 |      |                   |
+    +-------+  +-------+ +-------+  +-------+      |              +----+------+
+                                             +-----+------+       |br_int     |
+                                             |            |       |PHYSNET0   |
+                                      +------+-----++-----+------+|BW_EGR:1000|
+                                      |esn1        ||esn2        |+-----------+
+                                      |PHYSNET1    ||PHYSNET2    |
+                                      |BW_EGR:10000||BW_EGR:20000|
+                                      +------------++------------+
+
+
+                             +--------------------+
+                             | compute node (cn2) |
+                             | VCPU: 8 (3 used)   |
+                             | MEMORY_MB: 2048    |
+     >>....(from ss1)........|   min_unit: 1024   |
+                             |   step_size: 128   |
+                             | DISK_GB: 1000      |
+                             | traits: FOO        |
+                             | agg: [aggA]        |
+                             +---------+----------+
+                                       |
+            +--------------------------+----+---------------------------+
+            |                               |                           |
+      +-----+-----+                   +-----+-----+               +-----+-----+
+      |nic1       |                   |nic2       |               |nic3       |
+      |HW_NIC_ROOT|                   |HW_NIC_ROOT|               |HW_NIC_ROOT|
+      +-----+-----+                   +-----+-----+               +-----+-----+
+            |                               |                           |
+       +----+----+            +---------+---+-----+---------+           |
+       |         |            |         |         |         |           |
+    +--+--+   +--+--+      +--+--+   +--+--+   +--+--+   +--+--+     +--+--+
+    |pf1_1|   |pf1_2|      |pf2_1|   |pf2_2|   |pf2_3|   |pf2_4|     |pf3_1|
+    |NET1 |   |NET2 |      |NET1 |   |NET2 |   |NET1 |   |NET2 |     |NET1 |
+    |VF:4 |   |VF:4 |      |VF:2 |   |VF:2 |   |VF:2 |   |VF:2 |     |VF:8 |
+    +-----+   +-----+      +-----+   +-----+   +-----+   +-----+     +-----+
+    """
+
+    # Having these here allows us to pre-create cn1 and cn2 providers in
+    # DeepNUMANetworkFixture, where they have additional parents.
+    cn1 = None
+    cn2 = None
+
+    def start_fixture(self):
+        super(NUMANetworkFixture, self).start_fixture()
+
+        self.make_entities()
+
+    def make_entities(self):
+        aggA_uuid = uuidutils.generate_uuid()
+        os.environ['AGGA_UUID'] = aggA_uuid
+
+        ss1 = tb.create_provider(self.context, 'ss1', aggA_uuid)
+        tb.set_traits(ss1, ot.MISC_SHARES_VIA_AGGREGATE)
+        tb.add_inventory(ss1, orc.DISK_GB, 2000)
+        os.environ['SS1_UUID'] = ss1.uuid
+
+        # CN1
+        if not self.cn1:
+            self.cn1 = tb.create_provider(self.context, 'cn1', aggA_uuid)
+        self.cn1.set_aggregates([aggA_uuid])
+        tb.set_traits(self.cn1, ot.COMPUTE_VOLUME_MULTI_ATTACH)
+        os.environ['CN1_UUID'] = self.cn1.uuid
+
+        numas = []
+        for i in (0, 1):
+            numa = tb.create_provider(
+                self.context, 'numa%d' % i, parent=self.cn1.uuid)
+            traits = [ot.HW_NUMA_ROOT]
+            if i == 1:
+                traits.append('CUSTOM_FOO')
+            tb.set_traits(numa, *traits)
+            tb.add_inventory(numa, orc.VCPU, 4)
+            numas.append(numa)
+            os.environ['NUMA%d_UUID' % i] = numa.uuid
+        tb.add_inventory(
+            numas[0], orc.MEMORY_MB, 2048, min_unit=512, step_size=256)
+        tb.add_inventory(
+            numas[1], orc.MEMORY_MB, 2048, min_unit=256, max_unit=1024)
+        user, proj = tb.create_user_and_project(self.context, prefix='numafx')
+        consumer = tb.ensure_consumer(self.context, user, proj)
+        tb.set_allocation(self.context, numas[0], consumer, {orc.VCPU: 2})
+
+        fpga = tb.create_provider(self.context, 'fpga0', parent=numas[0].uuid)
+        # TODO(efried): Use standard FPGA resource class
+        tb.add_inventory(fpga, 'CUSTOM_FPGA', 1)
+        os.environ['FPGA0_UUID'] = fpga.uuid
+
+        pgpu = tb.create_provider(self.context, 'pgpu0', parent=numas[0].uuid)
+        tb.add_inventory(pgpu, orc.VGPU, 8)
+        os.environ['PGPU0_UUID'] = pgpu.uuid
+
+        for i in (0, 1):
+            fpga = tb.create_provider(
+                self.context, 'fpga1_%d' % i, parent=numas[1].uuid)
+            # TODO(efried): Use standard FPGA resource class
+            tb.add_inventory(fpga, 'CUSTOM_FPGA', 1)
+            os.environ['FPGA1_%d_UUID' % i] = fpga.uuid
+
+        agent = tb.create_provider(
+            self.context, 'sriov_agent', parent=self.cn1.uuid)
+        tb.set_traits(agent, 'CUSTOM_VNIC_TYPE_DIRECT')
+        os.environ['SRIOV_AGENT_UUID'] = agent.uuid
+
+        for i in (1, 2):
+            dev = tb.create_provider(
+                self.context, 'esn%d' % i, parent=agent.uuid)
+            tb.set_traits(dev, 'CUSTOM_PHYSNET%d' % i)
+            tb.add_inventory(dev, orc.NET_BW_EGR_KILOBIT_PER_SEC, 10000 * i)
+            os.environ['ESN%d_UUID' % i] = dev.uuid
+
+        agent = tb.create_provider(
+            self.context, 'ovs_agent', parent=self.cn1.uuid)
+        tb.set_traits(agent, 'CUSTOM_VNIC_TYPE_NORMAL')
+        os.environ['OVS_AGENT_UUID'] = agent.uuid
+
+        dev = tb.create_provider(self.context, 'br_int', parent=agent.uuid)
+        tb.set_traits(dev, 'CUSTOM_PHYSNET0')
+        tb.add_inventory(dev, orc.NET_BW_EGR_KILOBIT_PER_SEC, 1000)
+        os.environ['BR_INT_UUID'] = dev.uuid
+
+        # CN2
+        if not self.cn2:
+            self.cn2 = tb.create_provider(self.context, 'cn2')
+
+        self.cn2.set_aggregates([aggA_uuid])
+        tb.add_inventory(self.cn2, orc.VCPU, 8)
+        # Get a new consumer
+        consumer = tb.ensure_consumer(self.context, user, proj)
+        tb.set_allocation(self.context, self.cn2, consumer, {orc.VCPU: 3})
+        tb.add_inventory(
+            self.cn2, orc.MEMORY_MB, 2048, min_unit=1024, step_size=128)
+        tb.add_inventory(self.cn2, orc.DISK_GB, 1000)
+        tb.set_traits(self.cn2, 'CUSTOM_FOO')
+        os.environ['CN2_UUID'] = self.cn2.uuid
+
+        nics = []
+        for i in (1, 2, 3):
+            nic = tb.create_provider(
+                self.context, 'nic%d' % i, parent=self.cn2.uuid)
+            # TODO(efried): Use standard HW_NIC_ROOT trait
+            tb.set_traits(nic, 'CUSTOM_HW_NIC_ROOT')
+            nics.append(nic)
+            os.environ['NIC%s_UUID' % i] = nic.uuid
+        # PFs for NIC1
+        for i in (1, 2):
+            suf = '1_%d' % i
+            pf = tb.create_provider(
+                self.context, 'pf%s' % suf, parent=nics[0].uuid)
+            tb.set_traits(pf, 'CUSTOM_PHYSNET%d' % i)
+            # TODO(efried): Use standard generic VF resource class?
+            tb.add_inventory(pf, 'CUSTOM_VF', 4)
+            os.environ['PF%s_UUID' % suf] = pf.uuid
+        # PFs for NIC2
+        for i in (0, 1, 2, 3):
+            suf = '2_%d' % (i + 1)
+            pf = tb.create_provider(
+                self.context, 'pf%s' % suf, parent=nics[1].uuid)
+            tb.set_traits(pf, 'CUSTOM_PHYSNET%d' % ((i % 2) + 1))
+            # TODO(efried): Use standard generic VF resource class?
+            tb.add_inventory(pf, 'CUSTOM_VF', 2)
+            os.environ['PF%s_UUID' % suf] = pf.uuid
+        # PF for NIC3
+        suf = '3_1'
+        pf = tb.create_provider(
+            self.context, 'pf%s' % suf, parent=nics[2].uuid)
+        tb.set_traits(pf, 'CUSTOM_PHYSNET1')
+        # TODO(efried): Use standard generic VF resource class?
+        tb.add_inventory(pf, 'CUSTOM_VF', 8)
+        os.environ['PF%s_UUID' % suf] = pf.uuid
+
+
+class DeepNUMANetworkFixture(NUMANetworkFixture):
+    """Extend the NUMANetworkFixture with two empty resource providers as
+    parents and grandparents of the compute nodes. This is to exercise
+    same_subtree in a more complete fashion.
+    """
+    def make_entities(self):
+        """Create parents and grandparents for cn1 and cn2. They will be fully
+        populated by the superclass, NUMANetworkFixture.
+        """
+        grandparent1 = tb.create_provider(self.context, 'gp1')
+        parent1 = tb.create_provider(
+            self.context, 'p1', parent=grandparent1.uuid)
+        parent2 = tb.create_provider(
+            self.context, 'p2', parent=grandparent1.uuid)
+        self.cn1 = tb.create_provider(self.context, 'cn1', parent=parent1.uuid)
+        self.cn2 = tb.create_provider(self.context, 'cn2', parent=parent2.uuid)
+        super(DeepNUMANetworkFixture, self).make_entities()
+
+
 class NonSharedStorageFixture(APIFixture):
-    """An APIFixture that has two compute nodes with local storage that do not
-    use shared storage.
+    """An APIFixture that has three compute nodes with local storage that do
+    not use shared storage.
     """
     def start_fixture(self):
         super(NonSharedStorageFixture, self).start_fixture()
@@ -376,12 +618,14 @@ class NonSharedStorageFixture(APIFixture):
 
         cn1 = tb.create_provider(self.context, 'cn1')
         cn2 = tb.create_provider(self.context, 'cn2')
+        cn3 = tb.create_provider(self.context, 'cn3')
 
         os.environ['CN1_UUID'] = cn1.uuid
         os.environ['CN2_UUID'] = cn2.uuid
+        os.environ['CN3_UUID'] = cn3.uuid
 
         # Populate compute node inventory for VCPU, RAM and DISK
-        for cn in (cn1, cn2):
+        for cn in (cn1, cn2, cn3):
             tb.add_inventory(cn, 'VCPU', 24)
             tb.add_inventory(cn, 'MEMORY_MB', 128 * 1024)
             tb.add_inventory(cn, 'DISK_GB', 2000)
@@ -392,12 +636,13 @@ class CORSFixture(APIFixture):
 
     def start_fixture(self):
         super(CORSFixture, self).start_fixture()
-        # NOTE(cdent): If we remove this override, then the cors
-        # group ends up not existing in the conf, so when deploy.py
-        # wants to load the CORS middleware, it will not.
+        # Turn on the CORS middleware by setting 'allowed_origin'.
         self.conf_fixture.config(
             group='cors',
             allowed_origin='http://valid.example.com')
+        self.conf_fixture.config(
+            group='cors',
+            allow_headers=['openstack-api-version'])
 
 
 class GranularFixture(APIFixture):
@@ -504,10 +749,26 @@ class OpenPolicyFixture(APIFixture):
         for rule in policies.list_rules():
             name = rule.name
             # Ignore "base" rules for role:admin.
-            if name in ['placement', 'admin_api']:
+            if name in ('admin_api',):
                 continue
             rules[name] = '@'
         self.policy_fixture.set_rules(rules)
 
     def stop_fixture(self):
         super(OpenPolicyFixture, self).stop_fixture()
+
+
+class SecureRBACPolicyFixture(APIFixture):
+    """An APIFixture that enforce secure default policies and scope."""
+
+    _secure_rbac = True
+
+
+# Even though this just configures the defaults for enforce_scope and
+# enforce_new_default, it's useful because it's explicit in saying we're
+# testing old policy behavior. We can remove this once placement removes its
+# deprecated policies.
+class LegacyRBACPolicyFixture(APIFixture):
+    """An APIFixture that enforce deprecated policies."""
+
+    _secure_rbac = False
